/*******************************************************************************
* 
* Debug interface to P4Python
*
* Copyright (c) 2007-2014, Perforce Software, Inc.  All rights reserved.
* 
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions are met:
* 
* 1.  Redistributions of source code must retain the above copyright
*     notice, this list of conditions and the following disclaimer.
* 
* 2.  Redistributions in binary form must reproduce the above copyright
*     notice, this list of conditions and the following disclaimer in the
*     documentation and/or other materials provided with the distribution.
* 
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTR
* IBUTORS "AS IS"
* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
* ARE DISCLAIMED. IN NO EVENT SHALL PERFORCE SOFTWARE, INC. BE LIABLE FOR ANY
* DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
* (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
* ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
* (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
* SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*
* $Id: //depot/r22.1/p4-python/P4PythonDebug.h#1 $
*
*******************************************************************************/

#ifndef P4PYTHONDEBUG_H
# define P4PYTHONDEBUG_H

# define P4PYDBG_COMMANDS	1
# define P4PYDBG_CALLS		2
# define P4PYDBG_DATA		3
# define P4PYDBG_GC		4
# define P4PYDBG_RPC		9

class PythonDebug
{
private:
    int debugLevel;
    PyObject * logger;

public:
    PythonDebug();

    void setDebug(int d);
    int getDebug() const;

    void setLogger(PyObject * l);
    PyObject * getLogger () const;

    void debug    (int minLevel, const char * text);
    void info     (const char * text);
    void warning  (const char * text);
    void error    (const char * text);
    void critical (const char * text);

    void printDebug(int minLevel, const char * text);

private:
    void callLogger(const char * method, const char * text);
    void printDebug(const char * text);
};

#endif
