/*******************************************************************************

Copyright (c) 2007-2008, Perforce Software, Inc.  All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

1.  Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

2.  Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED. IN NO EVENT SHALL PERFORCE SOFTWARE, INC. BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*******************************************************************************/

/*******************************************************************************
 * Name		: undefdups.h
 *
 * Description	: Undefine portability macros defined by both Python's header
 * 		  files and those of the Perforce API. The idea is that you
 * 		  first include Python's headers, then this file, and then the
 * 		  Perforce API. This squelches any compiler warnings about
 * 		  pre-processor macros already being defined.
 *
 ******************************************************************************/

//
// Symbols defined by both Python and Perforce API headers
//
#undef HAVE_FSYNC
#undef HAVE_TRUNCATE
#undef HAVE_FORK
#undef HAVE_MMAP
#undef HAVE_UTIMES

#ifdef OS_NT

// Stupid "#define SetPort SetPortA" in winspool.h
#  ifdef SetPort
#    undef SetPort
#  endif

// GetMessage often #defined to GetMessageA
#  ifdef GetMessage
#    undef GetMessage
#  endif

#endif

// maintain backwards compatability with python 2.4
#if (PY_VERSION_HEX < 0x02050000)

typedef int Py_ssize_t;
#define PyErr_WarnEx(obj, str, size) PyErr_Warn(obj, str)

#endif
